#================================
# Mallows Model Averaging Program 
#================================
# Shahram M. Amini
# Virginia Tech
# shahram@vt.edu
# Last update: Feb, 23, 2012
#================================


# Loading libraries
library(foreign);library(xtable);library(LowRankQP)
#==================================================

            
#====================================================
# This chunk of code produces the top part of Table 1 
#====================================================
mydata <- read.dta("fls.dta", convert.dates=TRUE, convert.factors=TRUE, 
           missing.type=TRUE, convert.underscore=TRUE, warn.missing.labels=TRUE)

x.data <- mydata[,-1]
x <- sapply(1:ncol(x.data),function(i){x.data[,i]/max(x.data[,i])})   
scale.vector <- as.matrix(sapply(1:ncol(x.data),function(i){max(x.data[,i])}))        
Y <- as.matrix(mydata[,1])
output.colnames <- colnames(x.data)
full.fit <- lm(Y~x-1)
beta.full <- as.matrix(coef(full.fit))
M <- k <- ncol(x)
n <- nrow(x)
beta <- matrix(0,k,M)
e <- matrix(0,n,M)
K_vector <- matrix(c(1:M))
var.matrix <- matrix(0,k,M)
bias.sq <- matrix(0,k,M)             

# BMA estimator using BMS package
fls.bms <- bms(mydata[,-2], burn=100000, iter=1000000, g="BRIC", mprior="uniform", nmodel=2000, mcmc="bd", user.int=FALSE)
BMA.fls <- round(coef(fls.bms),4)[,1:3]

# MMA Estimator using orthogonalization 
for(i in 1:M)
           {
            X <- as.matrix(x[,1:i])
            ortho <- eigen(t(X)%*%X)
            Q <- ortho$vectors ; lambda <- ortho$values 
            x.tilda <- X%*%Q%*%(diag(lambda^-0.5,i,i))
            beta.star <- t(x.tilda)%*%Y
            beta.hat <- Q%*%diag(lambda^-0.5,i,i)%*%beta.star
            beta[1:i,i] <- beta.hat
            e[,i] <- Y-x.tilda%*%as.matrix(beta.star)
            bias.sq[,i] <- (beta[,i]-beta.full)^2
            var.matrix.star <- diag(as.numeric(((t(e[,i])%*%e[,i])/(n-i))),i,i)
            var.matrix.hat <- var.matrix.star%*%(Q%*%diag(lambda^-1,i,i)%*%t(Q))
            var.matrix[1:i,i] <- diag(var.matrix.hat)
            var.matrix[,i] <- var.matrix[,i]+ bias.sq[,i]
            } # End loop over i

             e_k <- e[,M]
             sigma_hat <- as.numeric((t(e_k)%*%e_k)/(n-M))
             G <- t(e)%*%e
             a <- ((sigma_hat)^2)*K_vector
             A <- matrix(1,1,M)
             b <- matrix(1,1,1)
             u <- matrix(1,M,1)
             optim <- LowRankQP(Vmat=G,dvec=a,Amat=A,bvec=b,uvec=u,method="LU",verbose=FALSE)
             weights <- as.matrix(optim$alpha)
             beta.scaled <- beta%*%weights
             final.beta <- beta.scaled/scale.vector
             std.scaled <- sqrt(var.matrix)%*%weights
             final.std <- std.scaled/scale.vector
             results.reduced <- as.matrix(cbind(final.beta,final.std))
             rownames(results.reduced) <- output.colnames; colnames(results.reduced) <- c("Coefficient", "Sd. Err")
             MMA.fls <- round(results.reduced,4)[-1,]
 
# WALS Estimator using Laplace distribution
x1 <- x[,1]; x2 <- x[,-1]; k1 <- ncol(as.matrix(x1)); k2 <- ncol(x2)
var1r <- solve(t(x1)%*%x1)
M1 <- diag(rep(1,n))- x1%*%solve(t(x1)%*%x1)%*%t(x1)
Z <- t(x2)%*%M1%*%x2
ortho <- eigen(Z)
P <- ortho$vectors ; lambda <- ortho$values 
x2star <- x2%*%P%*%(diag(lambda^-0.5,k2,k2))
xstar <- cbind(x1,x2star)
colnames(xstar) <- output.colnames
full.regress <- lm(Y~xstar-1)
beta.hat <- full.regress$coef
resid <- full.regress$resid
sigma <- sqrt((t(resid)%*% resid)/(n-k))
beta2star.hat <- beta.hat[-c(1:k1)] 
eta.hat <- beta2star.hat/sigma
c <- log(2)
signx <- sign(eta.hat)
xx <- abs(eta.hat)
g0 <- pnorm(-xx-c,0,1)
g1 <- pnorm(xx-c,0,1)
g2 <- dnorm(xx-c,0,1)
psi0 <- g0/g1
psi1 <- g2/g1
psi2 <- exp(2*c*xx)*psi0
hratio <- (1-psi2)/(1+psi2)
mu.tmp <- signx * (xx - c * hratio)
var.tmp <- 1 + (1-hratio^2) * c^2 - (1+hratio) * c * psi1
mu <- mu.tmp / lambda^0.5
var <- var.tmp / lambda
beta2 <- as.numeric(sigma) * P%*% as.matrix(mu)
var.beta2 <- as.numeric(sigma^2) * P%*% diag(var) %*% t(P)
Q <- var1r %*% t(x1) %*% x2
beta1 <- var1r %*% t(x1) %*% (Y - x2%*%beta2)
var.beta1 <- as.numeric(sigma^2)*var1r + Q %*% var.beta2 %*% t(Q)
beta.WALS <- as.matrix(c(beta1,beta2)/scale.vector)
se.WALS <- as.matrix(c(var.beta1^0.5, diag(var.beta2)^0.5)/scale.vector)
WALS.estimator <- round(cbind(beta.WALS,se.WALS),4)
colnames(WALS.estimator) <- c("Coefficient", "Std. Err")
rownames(WALS.estimator) <- output.colnames
WALS.fls <- WALS.estimator[-1,]
list(BMS.fls,WALS.fls,MMA.fls)
#============================= End of code for top part of Table 1 =============================



#=======================================================
# This chunk of code produces the middle part of Table 1 
#=======================================================
mydata <- read.dta("mp.dta", convert.dates=TRUE, convert.factors=TRUE, 
           missing.type=TRUE, convert.underscore=TRUE, warn.missing.labels=TRUE)

x.data <- mydata[,-1]
x <- sapply(1:ncol(x.data),function(i){x.data[,i]/max(x.data[,i])})   
scale.vector <- as.matrix(sapply(1:ncol(x.data),function(i){max(x.data[,i])}))        
Y <- as.matrix(mydata[,1])
output.colnames <- colnames(x.data)
full.fit <- lm(Y~x-1)
beta.full <- as.matrix(coef(full.fit))
M <- k <- ncol(x)
n <- nrow(x)
beta <- matrix(0,k,M)
e <- matrix(0,n,M)
K_vector <- matrix(c(1:M))
var.matrix <- matrix(0,k,M)
bias.sq <- matrix(0,k,M)             

# BMA estimator using BMS package
fls.bms <- bms(mydata[,-2], burn=100000, iter=1000000, g="BRIC", mprior="uniform", nmodel=2000, mcmc="bd", user.int=FALSE)
BMA.mp <- round(coef(fls.bms),4)[,1:3]

# MMA Estimator using orthogonalization 
for(i in 1:M)
           {
            X <- as.matrix(x[,1:i])
            ortho <- eigen(t(X)%*%X)
            Q <- ortho$vectors ; lambda <- ortho$values 
            x.tilda <- X%*%Q%*%(diag(lambda^-0.5,i,i))
            beta.star <- t(x.tilda)%*%Y
            beta.hat <- Q%*%diag(lambda^-0.5,i,i)%*%beta.star
            beta[1:i,i] <- beta.hat
            e[,i] <- Y-x.tilda%*%as.matrix(beta.star)
            bias.sq[,i] <- (beta[,i]-beta.full)^2
            var.matrix.star <- diag(as.numeric(((t(e[,i])%*%e[,i])/(n-i))),i,i)
            var.matrix.hat <- var.matrix.star%*%(Q%*%diag(lambda^-1,i,i)%*%t(Q))
            var.matrix[1:i,i] <- diag(var.matrix.hat)
            var.matrix[,i] <- var.matrix[,i]+ bias.sq[,i]
            } # End loop over i

             e_k <- e[,M]
             sigma_hat <- as.numeric((t(e_k)%*%e_k)/(n-M))
             G <- t(e)%*%e
             a <- ((sigma_hat)^2)*K_vector
             A <- matrix(1,1,M)
             b <- matrix(1,1,1)
             u <- matrix(1,M,1)
             optim <- LowRankQP(Vmat=G,dvec=a,Amat=A,bvec=b,uvec=u,method="LU",verbose=FALSE)
             weights <- as.matrix(optim$alpha)
             beta.scaled <- beta%*%weights
             final.beta <- beta.scaled/scale.vector
             std.scaled <- sqrt(var.matrix)%*%weights
             final.std <- std.scaled/scale.vector
             results.reduced <- as.matrix(cbind(final.beta,final.std))
             rownames(results.reduced) <- output.colnames; colnames(results.reduced) <- c("Coefficient", "Sd. Err")
             MMA.mp <- round(results.reduced,4)[-1,]
 
# WALS Estimator using Laplace distribution
x1 <- x[,1]; x2 <- x[,-1]; k1 <- ncol(as.matrix(x1)); k2 <- ncol(x2)
var1r <- solve(t(x1)%*%x1)
M1 <- diag(rep(1,n))- x1%*%solve(t(x1)%*%x1)%*%t(x1)
Z <- t(x2)%*%M1%*%x2
ortho <- eigen(Z)
P <- ortho$vectors ; lambda <- ortho$values 
x2star <- x2%*%P%*%(diag(lambda^-0.5,k2,k2))
xstar <- cbind(x1,x2star)
colnames(xstar) <- output.colnames
full.regress <- lm(Y~xstar-1)
beta.hat <- full.regress$coef
resid <- full.regress$resid
sigma <- sqrt((t(resid)%*% resid)/(n-k))
beta2star.hat <- beta.hat[-c(1:k1)] 
eta.hat <- beta2star.hat/sigma
c <- log(2)
signx <- sign(eta.hat)
xx <- abs(eta.hat)
g0 <- pnorm(-xx-c,0,1)
g1 <- pnorm(xx-c,0,1)
g2 <- dnorm(xx-c,0,1)
psi0 <- g0/g1
psi1 <- g2/g1
psi2 <- exp(2*c*xx)*psi0
hratio <- (1-psi2)/(1+psi2)
mu.tmp <- signx * (xx - c * hratio)
var.tmp <- 1 + (1-hratio^2) * c^2 - (1+hratio) * c * psi1
mu <- mu.tmp / lambda^0.5
var <- var.tmp / lambda
beta2 <- as.numeric(sigma) * P%*% as.matrix(mu)
var.beta2 <- as.numeric(sigma^2) * P%*% diag(var) %*% t(P)
Q <- var1r %*% t(x1) %*% x2
beta1 <- var1r %*% t(x1) %*% (Y - x2%*%beta2)
var.beta1 <- as.numeric(sigma^2)*var1r + Q %*% var.beta2 %*% t(Q)
beta.WALS <- as.matrix(c(beta1,beta2)/scale.vector)
se.WALS <- as.matrix(c(var.beta1^0.5, diag(var.beta2)^0.5)/scale.vector)
WALS.estimator <- round(cbind(beta.WALS,se.WALS),4)
colnames(WALS.estimator) <- c("Coefficient", "Std. Err")
rownames(WALS.estimator) <- output.colnames
WALS.mp <- WALS.estimator[-1,]
list(BMS.mp,WALS.mp,MMA.mp)
#============================= End of code for middle part of Table 1 ===========================



#=======================================================
# This chunk of code produces the bottom part of Table 1 
#=======================================================
mydata <- read.dta("dw.dta", convert.dates=TRUE, convert.factors=TRUE, 
           missing.type=TRUE, convert.underscore=TRUE, warn.missing.labels=TRUE)

x.data <- mydata[,-1]
x <- sapply(1:ncol(x.data),function(i){x.data[,i]/max(x.data[,i])})   
scale.vector <- as.matrix(sapply(1:ncol(x.data),function(i){max(x.data[,i])}))        
Y <- as.matrix(mydata[,1])
output.colnames <- colnames(x.data)
full.fit <- lm(Y~x-1)
beta.full <- as.matrix(coef(full.fit))
M <- k <- ncol(x)
n <- nrow(x)
beta <- matrix(0,k,M)
e <- matrix(0,n,M)
K_vector <- matrix(c(1:M))
var.matrix <- matrix(0,k,M)
bias.sq <- matrix(0,k,M)             

# BMA estimator using BMS package
fls.bms <- bms(mydata[,-2], burn=100000, iter=1000000, g="BRIC", mprior="uniform", nmodel=2000, mcmc="bd", user.int=FALSE)
BMA.dw <- round(coef(fls.bms),4)[,1:3]

# MMA Estimator using orthogonalization 
for(i in 1:M)
           {
            X <- as.matrix(x[,1:i])
            ortho <- eigen(t(X)%*%X)
            Q <- ortho$vectors ; lambda <- ortho$values 
            x.tilda <- X%*%Q%*%(diag(lambda^-0.5,i,i))
            beta.star <- t(x.tilda)%*%Y
            beta.hat <- Q%*%diag(lambda^-0.5,i,i)%*%beta.star
            beta[1:i,i] <- beta.hat
            e[,i] <- Y-x.tilda%*%as.matrix(beta.star)
            bias.sq[,i] <- (beta[,i]-beta.full)^2
            var.matrix.star <- diag(as.numeric(((t(e[,i])%*%e[,i])/(n-i))),i,i)
            var.matrix.hat <- var.matrix.star%*%(Q%*%diag(lambda^-1,i,i)%*%t(Q))
            var.matrix[1:i,i] <- diag(var.matrix.hat)
            var.matrix[,i] <- var.matrix[,i]+ bias.sq[,i]
            } # End loop over i

             e_k <- e[,M]
             sigma_hat <- as.numeric((t(e_k)%*%e_k)/(n-M))
             G <- t(e)%*%e
             a <- ((sigma_hat)^2)*K_vector
             A <- matrix(1,1,M)
             b <- matrix(1,1,1)
             u <- matrix(1,M,1)
             optim <- LowRankQP(Vmat=G,dvec=a,Amat=A,bvec=b,uvec=u,method="LU",verbose=FALSE)
             weights <- as.matrix(optim$alpha)
             beta.scaled <- beta%*%weights
             final.beta <- beta.scaled/scale.vector
             std.scaled <- sqrt(var.matrix)%*%weights
             final.std <- std.scaled/scale.vector
             results.reduced <- as.matrix(cbind(final.beta,final.std))
             rownames(results.reduced) <- output.colnames; colnames(results.reduced) <- c("Coefficient", "Sd. Err")
             MMA.dw <- round(results.reduced,4)[-1,]
 
# WALS Estimator using Laplace distribution
x1 <- x[,1]; x2 <- x[,-1]; k1 <- ncol(as.matrix(x1)); k2 <- ncol(x2)
var1r <- solve(t(x1)%*%x1)
M1 <- diag(rep(1,n))- x1%*%solve(t(x1)%*%x1)%*%t(x1)
Z <- t(x2)%*%M1%*%x2
ortho <- eigen(Z)
P <- ortho$vectors ; lambda <- ortho$values 
x2star <- x2%*%P%*%(diag(lambda^-0.5,k2,k2))
xstar <- cbind(x1,x2star)
colnames(xstar) <- output.colnames
full.regress <- lm(Y~xstar-1)
beta.hat <- full.regress$coef
resid <- full.regress$resid
sigma <- sqrt((t(resid)%*% resid)/(n-k))
beta2star.hat <- beta.hat[-c(1:k1)] 
eta.hat <- beta2star.hat/sigma
c <- log(2)
signx <- sign(eta.hat)
xx <- abs(eta.hat)
g0 <- pnorm(-xx-c,0,1)
g1 <- pnorm(xx-c,0,1)
g2 <- dnorm(xx-c,0,1)
psi0 <- g0/g1
psi1 <- g2/g1
psi2 <- exp(2*c*xx)*psi0
hratio <- (1-psi2)/(1+psi2)
mu.tmp <- signx * (xx - c * hratio)
var.tmp <- 1 + (1-hratio^2) * c^2 - (1+hratio) * c * psi1
mu <- mu.tmp / lambda^0.5
var <- var.tmp / lambda
beta2 <- as.numeric(sigma) * P%*% as.matrix(mu)
var.beta2 <- as.numeric(sigma^2) * P%*% diag(var) %*% t(P)
Q <- var1r %*% t(x1) %*% x2
beta1 <- var1r %*% t(x1) %*% (Y - x2%*%beta2)
var.beta1 <- as.numeric(sigma^2)*var1r + Q %*% var.beta2 %*% t(Q)
beta.WALS <- as.matrix(c(beta1,beta2)/scale.vector)
se.WALS <- as.matrix(c(var.beta1^0.5, diag(var.beta2)^0.5)/scale.vector)
WALS.estimator <- round(cbind(beta.WALS,se.WALS),4)
colnames(WALS.estimator) <- c("Coefficient", "Std. Err")
rownames(WALS.estimator) <- output.colnames
WALS.dw <- WALS.estimator[-1,]
list(BMS.dw,WALS.dw,MMA.dw)
#========================== End of code for bottom part of Table 1 ===========================


